import os
import json
import socket
import tempfile
from dateutil import parser

import requests

import docker
from docker.errors import ImageNotFound


DOCKERFILE_LOCAL_TEMPLATE = '''
FROM {{IMAGE}}

ADD . /tmp/app
WORKDIR /tmp/app
ENV SERVICE_TYPE {{SERVICE_TYPE}}
ENV HANDLER {{HANDLER}}
RUN if test -r requirements.txt; then pip install --no-cache-dir -r requirements.txt; fi
'''

LOCAL_TEST_LABEL_KEY = 'abeja-platform'
LOCAL_TEST_LABEL_VALUE = 'local'


class LocalModelHandler:
    def __init__(self):
        self._client = docker.from_env()
        self._setup_docker_cli()

    def _setup_docker_cli(self):
        """
        setup cli

        docker client will default to connecting to the following respectively.
        unix://var/run/docker.sock : linux / osx
        tcp://127.0.0.1:2376       : windows

        judge operation system using by os.name
        cf. https://docs.python.org/3.5/library/os.html#os.name

        :return:
        """
        if os.name == 'nt':
            base_url = 'tcp://127.0.0.1:2376'
        else:
            base_url = 'unix://var/run/docker.sock'
        self._docker_cli = docker.APIClient(base_url=base_url)

    def check_installation(self):
        try:
            return self._client.version()
        except requests.exceptions.ConnectionError:
            return None

    def pull_image(self, image, tag, stdout=None):
        """
        pull image if the image does not exist

        :param image:
        :param tag:
        :param stdout:
        :return:
        """
        if self._get_image(image, tag):
            return

        for output in self._docker_cli.pull(image, tag=tag, stream=True):
            if not stdout:
                continue
            try:
                for line in self._parse_stream(output):
                    line = line['status'].rstrip()
                    stdout(str(line))
            except:
                continue

    def check_rebuild_required(self):
        """
        return true if requirements.txt exists and
        the last update is later than that of the specified image.
        target_last_modified and image_last_modified are time in utc.

        :param image:
        :param tag:
        :return:
        """
        target = 'requirements.txt'
        if not os.path.exists(target):
            return False

        target_last_modified = os.stat(target).st_mtime
        image = self._find_built_image_id()
        if image is None:
            return True

        try:
            image_info = self._docker_cli.inspect_image(image)
            dt = parser.parse(image_info['Created'])
            image_last_modified = dt.timestamp()
        except ImageNotFound:
            return True

        if target_last_modified >= image_last_modified:
            return True
        return False

    def build_image(self, image, tag, handler, no_cache=False, stdout=None):
        """
        build docker image adding label to be able to find it

        :param image:
        :param tag:
        :param handler:
        :param no_cache:
        :param stdout:
        :return:
        """
        # skip build if use cache is enabled and built image exists.
        if self._find_built_image_id() and not no_cache:
            return

        dockerfile = self._generate_dockerfile(image, tag, handler)
        with tempfile.NamedTemporaryFile(mode='w+t') as f:
            f.write(dockerfile)
            f.seek(0)   # put file pointer back the initial position for being read
            for output in self._docker_cli.build(dockerfile=f.name,
                                                 path=os.getcwd(),
                                                 labels={LOCAL_TEST_LABEL_KEY: LOCAL_TEST_LABEL_VALUE}):
                if not stdout:
                    continue
                try:
                    for line in self._parse_stream(output):
                        line = line['stream'].rstrip()
                        stdout(str(line))
                except:
                    continue

    def create_local_server(self, handler):
        """
        run container in background as local api server
        assign dynamic port number for host port.

        :param handler:
        :return:
        """
        image_id = self._find_built_image_id()
        random_port = self._get_free_tcp_port()
        container = self._client.containers.run(image_id,
                                                remove=True,
                                                detach=True,
                                                working_dir='/tmp/app',
                                                environment=['SERVICE_TYPE=HTTP', 'HANDLER={}'.format(handler)],
                                                ports={'5000/tcp': random_port},
                                                volumes={'{}'.format(os.getcwd()): {'bind': '/tmp/app', 'mode': 'rw'}},
                                                command=['/bin/sh', '-c', 'abeja-model run'])
        return LocalServer(container, port=random_port)

    def _get_image(self, image, tag):
        try:
            return self._client.images.get("{}:{}".format(image, tag))
        except ImageNotFound:
            return None

    def _find_built_image_id(self):
        images = self._client.images.list(filters={'label': LOCAL_TEST_LABEL_KEY})
        if not images:
            return None
        # pick the first because images is descending order list
        _, image_id = images[0].short_id.split(':')
        return image_id

    def _generate_dockerfile(self, image, tag, handler):
        image = "{}:{}".format(image, tag)
        return DOCKERFILE_LOCAL_TEMPLATE \
            .replace('{{IMAGE}}', image) \
            .replace('{{SERVICE_TYPE}}', 'HTTP') \
            .replace('{{HANDLER}}', str(handler))

    def _parse_stream(self, output):
        if type(output) == bytes:
            # for py3
            output = output.decode('utf-8')
        lines = output.rstrip().split('\r\n')
        for line in lines:
            yield json.loads(line)

    def _get_free_tcp_port(self):
        """
        get free tcp port in dynamic port range
        which is defined by IANA
        cf. https://www.iana.org/assignments/service-names-port-numbers/service-names-port-numbers.xhtml

        :return:
        """
        tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        tcp.bind(('', 0))
        addr, port = tcp.getsockname()
        tcp.close()
        return port


class LocalServer:
    def __init__(self, container, port=None):
        self._container = container
        self.endpoint = 'http://localhost:{}'.format(port)

    def stop(self):
        # much faster to use kill, because stop container takes about 10 sec.
        # self._container.stop()
        try:
            self._container.kill()
        except docker.errors.NotFound:
            # ignore because the container does not exist anymore.
            return
